#include "stdafx.h"
#include "WavePlayer.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

////////////////////////////////////////////////////
list<WAVEHDR*> WavePlayer::BufferList;
CRITICAL_SECTION WavePlayer::CriticalSection = {0, 0, 0, 0, 0, 0};
VSTPlugin* WavePlayer::DataServer = 0;
WORD  WavePlayer::NumBuffers = 16;
DWORD  WavePlayer::BufferSize = 4096;
HWAVEOUT WavePlayer::OutputDevice = 0;
bool WavePlayer::KeepPlaying = false;
DWORD WavePlayer::PlayPosition;
DWORD WavePlayer::StartPosition;
DWORD WavePlayer::EndPosition;
bool WavePlayer::Loop = false;
char* WavePlayer::tmpCache;
////////////////////////////////////////////////////

WavePlayer::WavePlayer(VSTPlugin* pServer)
{
	
	StartPosition = 0;
	PlayPosition = 0;
	EndPosition = 0;

	OutputDevice = 0;
	Volume = 0xFFFF;
	Pitch = 0x00010000;
	IsOpened = false;
	IsPlaying = false;
	IsPaused = false;
	BufferCount = 0;
	LastError = 0;

	InitializeCriticalSection(&CriticalSection);
	DataServer = pServer;

}

WavePlayer::~WavePlayer()
{
	if (IsOpened) UnInit();
	DeleteCriticalSection(&CriticalSection);
}

bool WavePlayer::InitDevice(DWORD bSize, BYTE nBuffers, BYTE nChannels, DWORD SampleRate, BYTE BitsPerSample)
{
	WAVEHDR*	tmpBuffer;
	WAVEFORMATEX wfx = {WAVE_FORMAT_PCM,
						nChannels,
						SampleRate,
						nChannels*SampleRate*BitsPerSample/8,
#pragma warning ( push, 3 )
						nChannels*BitsPerSample/8,
#pragma warning ( pop )
						BitsPerSample,
						0};

	if (bSize) BufferSize = bSize;
	if (nBuffers) NumBuffers = nBuffers;

	BufferLength = BufferSize*1000/wfx.nAvgBytesPerSec;
	tmpCache = new char[BufferSize];
	
	LastError = waveOutOpen(&OutputDevice, WAVE_MAPPER, &wfx, (DWORD)&Callback, 0, CALLBACK_FUNCTION);
	if (LastError != MMSYSERR_NOERROR) return false;
	//else
	IsOpened = true;

	for (int i=0 ; i<NumBuffers; i++)
	{
#pragma warning (once  : 4706)
		if ( !(tmpBuffer = new WAVEHDR) ) return false;
		tmpBuffer->dwBufferLength = BufferSize;
		tmpBuffer->dwBytesRecorded= 0;
		tmpBuffer->dwFlags = 0;
		tmpBuffer->dwLoops = 0;
		tmpBuffer->dwUser = 0;
		if( !(tmpBuffer->lpData = (char*) new char[BufferSize]) ) return false;
		BufferList.push_back(tmpBuffer);
	}

	return true;
}

bool WavePlayer::StartPlaying(DWORD position, DWORD endPosition)
{
	DWORD rSize = 0;
	
	if (IsOpened && !KeepPlaying)
	{
		StartPosition = PlayPosition = position;
		EndPosition = endPosition;

		Pause();
		EnterCriticalSection(&CriticalSection);
		KeepPlaying = false;
		for (list<WAVEHDR*>::iterator i = BufferList.begin(); i !=BufferList.end(); ++i)
		{
			if ( !(DataServer->GetDataBlock(tmpCache, &rSize, PlayPosition) ) )
			{
				KeepPlaying = false;
				LeaveCriticalSection(&CriticalSection);
				return false;
			}
			CopyMemory((void*) (*i)->lpData, tmpCache, rSize << 2); //Znow przeliczam na bajty
			LastError = waveOutPrepareHeader(OutputDevice, (*i), sizeof(WAVEHDR));
			if (LastError != MMSYSERR_NOERROR)
			{
				KeepPlaying = false;
				LeaveCriticalSection(&CriticalSection);
				return false;
			}

			LastError = waveOutWrite(OutputDevice, (*i), sizeof(WAVEHDR));
			if (LastError != MMSYSERR_NOERROR)
			{
				KeepPlaying = false;
				LeaveCriticalSection(&CriticalSection);
				return false;
			}
			//cout << "Buffer at address " << *i << " sent." << "\n";

			if ( (PlayPosition += BufferSize >> 2) >= EndPosition)
			{
				if (Loop)
				{
					PlayPosition = StartPosition;
				}
				else
				{
					KeepPlaying = false;
					LeaveCriticalSection(&CriticalSection);
					return false;
				}
			}
		}
		LeaveCriticalSection(&CriticalSection);
		KeepPlaying = true;
		Resume();
		return true;
	}
	else return false;
}

void WavePlayer::Callback(HWAVEOUT hwo,
						  UINT uMsg,
						  DWORD dwInstance,
						  DWORD dwParam1,
						  DWORD dwParam2)
{
	DWORD rSize = 0;

	if ( (uMsg == WOM_DONE) && KeepPlaying )
	{
		EnterCriticalSection(&CriticalSection);

		if ( DataServer->GetDataBlock(tmpCache, &rSize, PlayPosition ) )
		{
			CopyMemory( ((WAVEHDR*)dwParam1)->lpData, tmpCache, rSize << 2);
			waveOutWrite( OutputDevice, (WAVEHDR*)dwParam1, sizeof(WAVEHDR) );
			if ( (PlayPosition += BufferSize >> 2) >= EndPosition)
			{
				if (Loop) PlayPosition = StartPosition;
				else KeepPlaying = false;
			}
		}
		else
		{
			KeepPlaying = false;
			LeaveCriticalSection(&CriticalSection);
			return;
		}

		LeaveCriticalSection(&CriticalSection);
	}
}

DWORD WavePlayer::GetVolume(void)
{
	DWORD tmpVolume;
	if (IsOpened)
	{
		LastError = waveOutGetVolume(OutputDevice, &tmpVolume);
		if (LastError != MMSYSERR_NOERROR) return 0;
		else return tmpVolume;
	}
	else return 0;
}

bool WavePlayer::SetVolume(DWORD Volume)
{
	if (IsOpened)
	{
		LastError = waveOutSetVolume(OutputDevice, Volume);
		if (LastError != MMSYSERR_NOERROR) return false;
		else return true;
	}
	else return false;
}

bool WavePlayer::Pause(void)
{
	if (IsOpened && KeepPlaying)
	{
		LastError = waveOutPause(OutputDevice);
		if (LastError != MMSYSERR_NOERROR) return false;
		else
		{
			IsPaused = true;
			return true;
		}
	}
	else return false;
}

bool WavePlayer::Resume(void)
{
	if (IsOpened && KeepPlaying)
	{
		LastError = waveOutRestart(OutputDevice);
		if (LastError != MMSYSERR_NOERROR) return false;
		else
		{
			IsPaused = false;
			return true;
		}
	}
	else return false;
}

bool WavePlayer::isPaused(void)
{
	return IsPaused;
}

bool WavePlayer::SetPitch(double Pitch)
{
	DWORD hiWord = 0, loWord = 0;
	
	double integer = 0, fractional = 0;
	
	if ( (Pitch <= 0) || !IsOpened) return false;
	else
	{
		//This turns float into fixed-point value.
		fractional = modf(Pitch, &integer);
		hiWord = integer;
		loWord = fractional*65536;
		(hiWord << 16);
		hiWord += loWord;
		LastError = waveOutSetPlaybackRate(OutputDevice, hiWord);
		if (LastError !=MMSYSERR_NOERROR) return false;
		else return true;
	}
}

bool WavePlayer::StopPlaying(void)
{
	if (IsOpened)
	{
		KeepPlaying = false;
		LastError = waveOutReset(OutputDevice);
		if (LastError != MMSYSERR_NOERROR) return false;
		else return true;
	}
	else return false;
}

MMRESULT WavePlayer::_GetLastError(void)
{
	return LastError;
}

bool WavePlayer::UnInit(void)
{
	WAVEHDR*	tmpBuffer;

	if (KeepPlaying) StopPlaying();
	
	for (int i=0 ; i<NumBuffers; i++)
	{
		tmpBuffer = BufferList.front();
		waveOutUnprepareHeader( OutputDevice, tmpBuffer, sizeof(WAVEHDR) );
		delete tmpBuffer->lpData;
		delete tmpBuffer;
		BufferList.pop_front();
	}

	delete tmpCache;

	LastError = waveOutClose(OutputDevice);
	if (LastError != MMSYSERR_NOERROR) return false;
	else
	{
		IsOpened = false;
		return true;
	}
}		

void WavePlayer::SetLooping(bool doLoop)
{
	Loop = doLoop;
}

bool WavePlayer::_isPlaying(void)
{
	return KeepPlaying;
}